## By Erik Hintz, Marius Hofert, Christiane Lemieux


library(nvmix)
library(RColorBrewer)
library(copula)


## Sampling the $t$, grouped $t$ and skew $t$ copula with `nvmix`

str(rgStudentcopula)

### Sample 4-dimensional grouped t copula

set.seed(1) # set a seed for reproducibility
d <- 4 # dimension
groupings <- c(1, 1, 2, 2) # (U1, U2) and (U3, U4) are each in the same group
df <- c(0.5, 25) # degrees-of-freedom for groups 1 and 2
rho <- 0.7 # off-diagonal entry of P
scale <- matrix(rho, ncol = d, nrow = d)
diag(scale) <- 1 # scale is now P
n <- 1e3 # sample size
U <- rgStudentcopula(n, groupings = groupings, df = df, scale = scale)
pairs2(U, pch = ".")

### Estimate C(u,...,u) for a 4-dimensional grouped t copula

MC_est <- function(n, u) {
   le.u <- apply(rgStudentcopula(n, groupings = groupings, df = df,
                                 scale = scale) <= u, 1, all)
   c(est = mean(le.u), abs.err = 3.5 * sqrt(var(le.u)/n))
}

RQMC_est <- function(n, u, B = 15) {
   le.u <- sapply(1:B, function(i) mean(apply(rgStudentcopula(n, df = df,
                                                              groupings = groupings, scale = scale, method = "sobol") <= u, 1, all)))
   c(est = mean(le.u), abs.err = 3.5 * sqrt(var(le.u)/B))
}

u <- 0.5 # evaluation point
n <- 2^10 # sample size
B <- 15 # number of replications
set.seed(2) # for reproducibility
t.MC   <- system.time(MC.out <- MC_est(B * n, u = u))
t.RQMC <- system.time(RQMC.out <- RQMC_est(n, u = u, B = B))
matrix(c(MC.out, RQMC.out), ncol = 2, dimnames = list(c("est", "abs.err"),
                                                      c("MC", "RQMC")))
## Compare with pgstudentcopula()
t.pgstcop <- system.time(p <- pgStudentcopula(rep(u, d), groupings = groupings,
                                              df = df, scale = scale))
print(p)

## Run-times
c("MC" = t.MC[3], "RQMC" = t.RQMC[3], "pgStudentcop" = t.pgstcop[3])

### ... now for several 'u'
u <- seq(from = 0.003, to = 0.2, length.out = 27)
MC.out  <- sapply(u, function(u.) MC_est(B * n, u = u.)[1])
RQMC.out <- sapply(u, function(u.) RQMC_est(n,  u = u., B = B)[1])
pgstcop.out <- pgStudentcopula(matrix(u, ncol = d, nrow = 27),
                               groupings = groupings, df = df, scale = scale)

## Plot
pal <- colorRampPalette(c("#000000", brewer.pal(8, name = "Dark2")[c(7, 3, 5)]))
palette(pal(3))
plot(NA, xlim = range(u), ylim = range(MC.out, RQMC.out, pgstcop.out), xlab = "u",
     ylab = "C(u,..,u)", log = "y")
lines(u, pgstcop.out, col = 1)
lines(u, RQMC.out, col = 2)
lines(u, MC.out, col = 3)
legend("topleft", c("pgStudentcopula()", "RQMC.est()", "MC.est()"), lty = rep(1, 3),
       col = 1:3, bty = "n")

### Sampling skew-t copulas: pobs = TRUE vs pobs = FALSE

n <- 1000; d <- 10; rho <- 0.5; gamma <- rep(1, d); df <- 7
scale <- matrix(rho, ncol = d, nrow = d); diag(scale) <- 1
set.seed(1)
system.time(samplecop.pobs <- rskewtcopula(n, scale = scale, gamma = gamma,
                                           df = df, pseudo = TRUE))
set.seed(1)
system.time(samplecop.pskewt <- rskewtcopula(n, scale = scale, gamma = gamma,
                                             df = df, pseudo = FALSE))

## Plot
opar <- par(no.readonly = TRUE)
par(pty = "s")
layout(rbind(1:2))
plot(samplecop.pobs, xlab = expression(U[1]), ylab = expression(U[2]))
mtext("pobs = TRUE", side = 4)
plot(samplecop.pskewt, xlab = expression(U[1]), ylab = expression(U[2]))
mtext("pobs = FALSE", side = 4)
layout(1)
par(opar)

## Compute estimates and record run-time when estimating the probability
## based on pseudo = TRUE and pseudo = FALSE
ns <- c(50, 100, 250, 500, 750, 1000, 2000, 3000, 4000, 5000) # sample sizes
res <- array(NA, dim = c(length(ns), 2, 2),
             dimnames = list(n = ns, pobs = c(TRUE, FALSE), c("est", "CPU")))
for(i in seq_along(ns)) {
   set.seed(12) # same seed for all sample sizes
   res[i, 1, 2] <- system.time(res[i, 1, 1] <- mean(apply(
      rskewtcopula(ns[i], scale = scale, gamma = gamma, df = df, pseudo = TRUE),
      1, function(i) all(i > 0.9))))[1]
   set.seed(12) # same random numbers for 'pseudo = TRUE' and 'pseudo = FALSE'
   res[i, 2, 2] <- system.time(res[i, 2, 1] <- mean(apply(
      rskewtcopula(ns[i], scale = scale, gamma = gamma, df = df, pseudo = FALSE),
      1, function(i) all(i > 0.9))))[1]
}

## Plot
opar <- par(no.readonly = TRUE)
par(pty = "s")
palette(pal(2))
layout(rbind(1:2))
## Estimates
plot(NA, xlim = range(ns), ylim = range(res[,, 1]), log = "y",
     xlab = "n", ylab = "Probability")
for(i in 1:2) lines(ns, res[,i,1], col = i)
legend("bottomright", c("pobs = TRUE", "pobs = FALSE"), lty = rep(1, 2),
       col = 1:2, bty = 'n')
## CPU
plot(NA, xlim = range(ns), ylim = range(res[,, 2]),
     xlab = "n", ylab = "CPU", log = "y")
for(i in 1:2) lines(ns, res[,i,2], col = i)
legend("bottomright", c("pobs = TRUE", "pobs = FALSE"), lty = rep(1, 2),
       col = 1:2, bty = 'n')
layout(1)
par(opar)


## Estimating the $t$, grouped $t$ and skew $t$ copula with `nvmix`

### Example call for the t copula
n <- 500; d <- 7; rho <- 0.5; df <- 3.5
scale <- matrix(rho, ncol = d, nrow = d); diag(scale) <- 1
set.seed(123)
sample <- rStudentcopula(n, df = df, scale = scale)
fitStudentcopula(sample)


### Comparing 4 methods to fit the t copula


set.seed(1)
methods <- c("Moment-MLE", "EM-MLE", "Full-MLE", "Full-MLE ('copula')")
reps <- 50 # number of replications for each method
res <- array(, dim = c(length(methods), reps, 4),
             dimnames = list(method = methods, rep = 1:reps,
                             c("loglik", "cpu", "df", "rho12")))
for(i in 1:reps) {
   U <- rStudentcopula(n, df = df, scale = scale) # sample
   t.mm   <-  system.time(fit.mm  <- fitStudentcopula(U, fit.method = "Moment-MLE"))[1]
   t.em   <-  system.time(fit.em  <- fitStudentcopula(U, fit.method = "EM-MLE"))[1]
   t.fl <- system.time(fit.fl <- fitStudentcopula(U, fit.method = "Full-MLE"))[1]
   t.fl.cop <- system.time(fit.fl.cop <- fitCopula(
      tCopula(dim = d, dispstr = "un"), U, method = "ml", estimate.variance = FALSE,
      start = c(fit.mm$scale[upper.tri(fit.mm$scale)], 5)))[1]
   res[, i, "loglik"] <- c(fit.mm$max.ll, fit.em$max.ll, fit.fl$max.ll,
                           fit.fl.cop@loglik)
   res[, i, "cpu"] <- c(t.mm, t.em, t.fl, t.fl.cop)
   res[, i, "df"] <- c(fit.mm$df, fit.em$df, fit.fl$df,
                       fit.fl.cop@estimate[d*(d-1)/2+1])
   res[, i, "rho12"] <- c(fit.mm$scale[1, 2], fit.em$scale[1, 2], fit.fl$scale[1, 2],
                          fit.fl.cop@estimate[1])
}

## Plot
opar <- par(no.readonly = TRUE)
par(pty = "s")
abserr.nu <- abs(res[, , "df"] - df)
abserr.rho12 <- abs(res[,, "rho12"] - scale[1, 2])
layout(rbind(1:3))
boxplot(t(abserr.nu), ylab = expression(abs(hat(nu)-nu)))
boxplot(t(abserr.rho12), ylab = expression(abs(hat(rho)[12]-rho[12])))
boxplot(t(res[,, "cpu"]), ylab = "Run-times")
layout(1)
par(opar)

### Comparing two methods for the grouped t copula

ns <- c(50, 250, 500, 750, 1000) # sample sizes
reps <- 10 # number of repetitions for each sample size
d <- 4 # dimension
df <- c(3, 8) # degrees-of-freedom for each group
grp <- rep(1:2, each = 2) # 2 components in each group
set.seed(1)
scale <- cov2cor(rWishart(1, d, diag(d))[,,1]) # same known scale for all reps
fit.res <- array(, dim = c(length(ns), reps, length(df), 2),
                 dimnames = list(n = ns, rep = 1:reps, df = c("df1", "df2"),
                                 est = c("init", "MLE")))
for(j in 1:reps) {
   set.seed(j)
   sample <- rgStudentcopula(max(ns), groupings = groupings, scale = scale, df = df)
   for(i in seq_along(ns)){
      fit <- fitgStudentcopula(u = sample[1:ns[i], ], groupings = grp,
                               scale = scale, verbose = FALSE)
      fit.res[i, j, , "init"] <- fit$df.init
      fit.res[i, j, , "MLE"]  <- fit$df
   }
}

## Plot

opar <- par(no.readonly = TRUE)
par(pty = "s")
layout(rbind(1:2))
## Create plot (initial estimates)
pal <- colorRampPalette(c("#000000", brewer.pal(8, name = "Dark2")[c(7, 3, 5)]))
palette(pal(2))
plot(NA, xlim = range(ns), ylim = range(fit.res[,,, ]), xlab = "n",
     ylab = "Initial estimate", xaxt = "n")
for(i in 1:reps){
   for(j in seq_along(df)){
      points(ns, fit.res[, i, j, "init"], col = j, pch = j)
   }
}
axis(1, at = ns)
abline(h = df, col = 1:2) # horizontal line with true df
mtext(side = 4,  "grouped t, d=4, 10 repetitions for each n", adj = 0)
legend("topright", c("Estimated df1", "Estimated df2", "True df1", "True df2"),
       col = rep(1:2, 2), pch = c(1:2, NA, NA), lty = c(NA, NA, 1, 1),
       bty = 'n')
## Create plot (MLEs)
plot(NA, xlim = range(ns), ylim = range(fit.res[,,, ]), xlab = "n", xaxt = "n",
     ylab = "MLEs")
for(i in 1:reps){
   for(j in seq_along(df)){
      points(ns, fit.res[, i, j, "MLE"], col = j, pch = j)
   }
}
axis(1, at = ns)
abline(h = df, col = 1:2) # horizontal line with true df
mtext(side = 4,  "grouped t, d=4, 10 repetitions for each n", adj = 0)
legend("topright", c("Estimated df1", "Estimated df2", "True df1", "True df2"),
       col = rep(1:2, 2), pch = c(1:2, NA, NA), lty = c(NA, NA, 1, 1),
       bty = 'n')
layout(1)
par(opar)

### Illustrate verbose

set.seed(42)
n <- 100; d <- 6; groupings <- rep(1:3, each = 2); df <- c(1, 4, 8)
rho <- runif(1, -1, 1)
P <- matrix(rho, ncol = d, nrow = d); diag(P) <- 1
U <- rgStudentcopula(n, groupings = groupings, df = df, scale = P)
(fit <- fitgStudentcopula(u = U, groupings = groupings))

## Now with larger 'maxit' for optim()
(fit <- fitgStudentcopula(u = U, groupings = groupings,
                          control = list(control.optim = list(maxit = 200))))


### Illustrating fitskewtcopula()

## Contour plot of profile log-lik
set.seed(12)
n <- 1000; d <- 3; rho <- 0.5; gamma <- rep(0.8, d); df <- 10
scale <- matrix(rho, ncol = d, nrow = d); diag(scale) <- 1
U <- rskewtcopula(n, scale = scale, gamma = gamma, df = df, pseudo = FALSE)
## Define log-likelihood function with known 'scale'
loglik <- function(par) sum(dskewtcopula(
   U, scale = scale, df = par[1], gamma = rep(par[2], d), log = TRUE))
## Evaluate 'loglik' on a grid and plot
n.grid <- 19
df_ <- seq(5, 11, length.out = n.grid)
gam_ <- seq(0, 2, length.out = n.grid)
grid <- expand.grid("nu" = df_, "gamma" = gam_)
lik <- apply(grid, 1, loglik)
val.lik <- cbind(grid, "loglik" = lik)
contourplot2(val.lik)

## Illustrate 'Full-MLE' vs 'EM-MLE'

t.em <- system.time(fit.em <- nvmix:::fitskewtcopula(
   U, fit.method = "EM-MLE", df.bounds = c(5, 15), gamma.bounds = c(0, 2)))[3]
t.fl <- system.time(fit.fl<- nvmix:::fitskewtcopula(
   U, fit.method = "Full-MLE", df.bounds = c(5, 15), gamma.bounds = c(0, 2)))[3]
t.em.neld <- system.time(fit.em.neld <- nvmix:::fitskewtcopula(
   U, fit.method = "EM-MLE", df.bounds = c(5, 15), gamma.bounds = c(0, 2),
   optim.method = "Nelder-Mead"))[3]
t.fl.neld <- system.time(fit.fl.neld<- nvmix:::fitskewtcopula(
   U, fit.method = "Full-MLE", df.bounds = c(5, 15), gamma.bounds = c(0, 2),
   optim.method = "Nelder-Mead"))[3]

## Print results

mat <- matrix(, ncol = 4, nrow = 7, dimnames = list(
   c("max.ll", "df", "gamma", "rho[1,2]","rho[1,3]", "rho[2,3]", "CPU"),
   c("EM (L-BFGS-B)", "Full (L-BFGS-B)", "EM (Nelder-Mead)", "Full (Nelder-Mead)")))
mat[, 1] <- c(fit.em$ll, fit.em$df, fit.em$gamma, fit.em$scale[1, 2],
              fit.em$scale[1, 3], fit.em$scale[2, 3], t.em)
mat[, 2] <- c(fit.fl$ll, fit.fl$df, fit.fl$gamma, fit.fl$scale[1, 2],
              fit.fl$scale[1, 3], fit.fl$scale[2, 3], t.fl)
mat[, 3] <- c(fit.em.neld$ll, fit.em.neld$df, fit.em.neld$gamma, fit.em.neld$scale[1, 2],
              fit.em.neld$scale[1, 3], fit.em.neld$scale[2, 3], t.em.neld)
mat[, 4] <- c(fit.fl.neld$ll, fit.fl.neld$df, fit.fl.neld$gamma, fit.fl.neld$scale[1, 2],
              fit.fl.neld$scale[1, 3], fit.fl.neld$scale[2, 3], t.fl.neld)
mat